#include "globals.h"
#include "../dynarec/dynarec.h"
#include "regcache.h"
#include "r4300i.h"

RETURN_TYPE rc_SPECIAL_instr(OP_PARAMS);
RETURN_TYPE rc_REGIMM_instr(OP_PARAMS);
RETURN_TYPE rc_COP0_instr(OP_PARAMS);
RETURN_TYPE rc_COP1_instr(OP_PARAMS);
RETURN_TYPE rc_TLB_instr(OP_PARAMS);
RETURN_TYPE rc_COP1_BC_instr(OP_PARAMS);
RETURN_TYPE rc_COP1_S_instr(OP_PARAMS);
RETURN_TYPE rc_COP1_D_instr(OP_PARAMS);
RETURN_TYPE rc_COP1_W_instr(OP_PARAMS);
RETURN_TYPE rc_COP1_L_instr(OP_PARAMS);
RETURN_TYPE rc_UNUSED(OP_PARAMS);

//---------------------------------------------------------------------------------------

/*
****************************************************************************
** Main CPU                                                               **
****************************************************************************
  
    CPU: Instructions encoded by opcode field.
    31---------26---------------------------------------------------0
    |  opcode   |                                                   |
    ------6----------------------------------------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 | *1    | *2    | J     | JAL   | BEQ   | BNE   | BLEZ  | BGTZ  |
001 | ADDI  | ADDIU | SLTI  | SLTIU | ANDI  | ORI   | XORI  | LUI   |
010 | *3    | *4    |  ---  |  ---  | BEQL  | BNEL  | BLEZL | BGTZL |
011 | DADDI |DADDIU | LDL   | LDR   |  ---  |  ---  |  ---  |  ---  |
100 | LB    | LH    | LWL   | LW    | LBU   | LHU   | LWR   | LWU   |
101 | SB    | SH    | SWL   | SW    | SDL   | SDR   | SWR   | CACHE |
110 | LL    | LWC1  |  ---  |  ---  | LLD   | LDC1  | (LDC2)| LD    |
111 | SC    | SWC1  |  ---  |  ---  | SCD   | SDC1  | (SDC2)| SD    |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
     *1 = SPECIAL, see SPECIAL list    *2 = REGIMM, see REGIMM list
     *3 = COP0                         *4 = COP1
*/
RETURN_TYPE (*rc_CPU_instruction[64])(OP_PARAMS)=
{
    rc_SPECIAL_instr,   rc_REGIMM_instr,    rc_r4300i_j,    rc_r4300i_jal,      rc_r4300i_beq,  rc_r4300i_bne,  rc_r4300i_blez, rc_r4300i_bgtz,
    rc_r4300i_addi, rc_r4300i_addiu,    rc_r4300i_slti,rc_r4300i_sltiu, rc_r4300i_andi,rc_r4300i_ori,   rc_r4300i_xori, rc_r4300i_lui,
    rc_COP0_instr,      rc_COP1_instr,      rc_UNUSED,      rc_UNUSED,          rc_r4300i_beql,rc_r4300i_bnel,rc_r4300i_blezl,  rc_r4300i_bgtzl,
    rc_r4300i_daddi,    rc_r4300i_daddiu,   rc_r4300i_ldl,  rc_r4300i_ldr,      rc_UNUSED,      rc_UNUSED,      rc_UNUSED,          rc_UNUSED,
    rc_r4300i_lb,       rc_r4300i_lh,       rc_r4300i_lwl,  rc_r4300i_lw,       rc_r4300i_lbu,  rc_r4300i_lhu,  rc_r4300i_lwr,      rc_r4300i_lwu,
    rc_r4300i_sb,       rc_r4300i_sh,       rc_r4300i_swl,  rc_r4300i_sw,       rc_r4300i_sdl,  rc_r4300i_sdr,  rc_r4300i_swr,      rc_r4300i_cache,
    rc_r4300i_ll,       rc_r4300i_lwc1, rc_UNUSED,      rc_UNUSED,          rc_r4300i_lld,  rc_r4300i_ldc1,rc_UNUSED,           rc_r4300i_ld,
    rc_r4300i_sc,       rc_r4300i_swc1, rc_UNUSED,      rc_UNUSED,          rc_r4300i_scd,  rc_r4300i_sdc1,rc_UNUSED,           rc_r4300i_sd
};

//---------------------------------------------------------------------------------------

/*
    SPECIAL: Instr. encoded by function field when opcode field = SPECIAL.
    31---------26------------------------------------------5--------0
    | = SPECIAL |                                         | function|
    ------6----------------------------------------------------6-----
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 | SLL   |  ---  | SRL   | SRA   | SLLV  |  ---  | SRLV  | SRAV  |
001 | JR    | JALR  |  ---  |  ---  |SYSCALL| BREAK |  ---  | SYNC  |
010 | MFHI  | MTHI  | MFLO  | MTLO  | DSLLV |  ---  | DSRLV | DSRAV |
011 | MULT  | MULTU | DIV   | DIVU  | DMULT | DMULTU| DDIV  | DDIVU |
100 | ADD   | ADDU  | SUB   | SUBU  | AND   | OR    | XOR   | NOR   |
101 |  ---  |  ---  | SLT   | SLTU  | DADD  | DADDU | DSUB  | DSUBU |
110 | TGE   | TGEU  | TLT   | TLTU  | TEQ   |  ---  | TNE   |  ---  |
111 | DSLL  |  ---  | DSRL  | DSRA  |DSLL32 |  ---  |DSRL32 |DSRA32 |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
*/
RETURN_TYPE (*rc_SPECIAL_Instruction[64])(OP_PARAMS)=
{
    rc_r4300i_sll,  rc_UNUSED,          rc_r4300i_srl,  rc_r4300i_sra,  rc_r4300i_sllv, rc_UNUSED,          rc_r4300i_srlv, rc_r4300i_srav,
    rc_r4300i_jr,   rc_r4300i_jalr, rc_UNUSED,      rc_UNUSED,      rc_r4300i_syscall,  rc_r4300i_break,    rc_UNUSED,          rc_r4300i_sync,
    rc_r4300i_mfhi,rc_r4300i_mthi,  rc_r4300i_mflo,rc_r4300i_mtlo,rc_r4300i_dsllv,  rc_UNUSED,          rc_r4300i_dsrlv,    rc_r4300i_dsrav,
    rc_r4300i_mult,rc_r4300i_multu, rc_r4300i_div,  rc_r4300i_divu,rc_r4300i_dmult, rc_r4300i_dmultu,   rc_r4300i_ddiv, rc_r4300i_ddivu,    
    rc_r4300i_add,  rc_r4300i_addu, rc_r4300i_sub,  rc_r4300i_subu,rc_r4300i_and,       rc_r4300i_or,       rc_r4300i_xor,      rc_r4300i_nor,
    rc_UNUSED,      rc_UNUSED,          rc_r4300i_slt,  rc_r4300i_sltu,rc_r4300i_dadd,  rc_r4300i_daddu,    rc_r4300i_dsub, rc_r4300i_dsubu,
    rc_r4300i_tge,  rc_r4300i_tgeu, rc_r4300i_tlt,  rc_r4300i_tltu,rc_r4300i_teq,       rc_UNUSED,          rc_r4300i_tne,      rc_UNUSED,
    rc_r4300i_dsll,rc_UNUSED,           rc_r4300i_dsrl,rc_r4300i_dsra,rc_r4300i_dsll32, rc_UNUSED,          rc_r4300i_dsrl32,   rc_r4300i_dsra32
};

//---------------------------------------------------------------------------------------

/*
    REGIMM: Instructions encoded by the rt field when opcode field = REGIMM.
    31---------26----------20-------16------------------------------0
    | = REGIMM  |          |   rt    |                              |
    ------6---------------------5------------------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
 00 | BLTZ  | BGEZ  | BLTZL | BGEZL |  ---  |  ---  |  ---  |  ---  |
 01 | TGEI  | TGEIU | TLTI  | TLTIU | TEQI  |  ---  | TNEI  |  ---  |
 10 | BLTZAL| BGEZAL|BLTZALL|BGEZALL|  ---  |  ---  |  ---  |  ---  |
 11 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
*/
RETURN_TYPE (*rc_REGIMM_Instruction[32])(OP_PARAMS)=
{
    rc_r4300i_bltz, rc_r4300i_bgez, rc_r4300i_bltzl,    rc_r4300i_bgezl,    rc_UNUSED,          rc_UNUSED,  rc_UNUSED,          rc_UNUSED,
    rc_r4300i_tgei, rc_r4300i_tgeiu,    rc_r4300i_tlti, rc_r4300i_tltiu,    rc_r4300i_teqi, rc_UNUSED,  rc_r4300i_tnei, rc_UNUSED,
    rc_r4300i_bltzal,   rc_r4300i_bgezal,   rc_r4300i_bltzall,  rc_r4300i_bgezall,  rc_UNUSED,          rc_UNUSED,  rc_UNUSED,          rc_UNUSED,
    rc_UNUSED,          rc_UNUSED,          rc_UNUSED,          rc_UNUSED,          rc_UNUSED,          rc_UNUSED,  rc_UNUSED,          rc_UNUSED   
};

//---------------------------------------------------------------------------------------

/*
****************************************************************************
** COP0                                                                   **
****************************************************************************

    COP0: Instructions encoded by the fmt field when opcode = COP0.
    31--------26-25------21 ----------------------------------------0
    |  = COP0   |   fmt   |                                         |
    ------6----------5-----------------------------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
 00 | MFC0  |  ---  |  ---  |  ---  | MTC0  |  ---  |  ---  |  ---  |
 01 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 10 | *1    |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 11 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
     *1 = TLB instr, see TLB list
*/
RETURN_TYPE (*rc_COP0_Instruction[32])(OP_PARAMS)=
{
    rc_r4300i_COP0_mfc0,    rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_r4300i_COP0_mtc0,    rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  
    rc_TLB_instr,           rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED
};

//---------------------------------------------------------------------------------------

/*
    TLB: Instructions encoded by the function field when opcode
         = COP0 and fmt = TLB.
    31--------26-25------21 -------------------------------5--------0
    |  = COP0   |  = TLB  |                               | function|
    ------6----------5-----------------------------------------6-----
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 |  ---  | TLBR  | TLBWI |  ---  |  ---  |  ---  | TLBWR |  ---  |
001 | TLBP  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
010 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
011 | ERET  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
100 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
101 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
110 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
111 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
*/
RETURN_TYPE (*rc_TLB_Instruction[64])(OP_PARAMS)=
{
    rc_UNUSED,                      rc_r4300i_COP0_tlbr,    rc_r4300i_COP0_tlbwi,   rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_r4300i_COP0_tlbwr,   rc_UNUSED,
    rc_r4300i_COP0_tlbp,            rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,                      rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,              rc_UNUSED,
    rc_r4300i_COP0_eret,            rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,                      rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,                      rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,                      rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,                      rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,              rc_UNUSED
};

//---------------------------------------------------------------------------------------

/*
****************************************************************************
** COP1 - Floating Point Unit (FPU)                                       **
****************************************************************************

    COP1: Instructions encoded by the fmt field when opcode = COP1.
    31--------26-25------21 ----------------------------------------0
    |  = COP1   |   fmt   |                                         |
    ------6----------5-----------------------------------------------
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
 00 | MFC1  | DMFC1 | CFC1  |  ---  | MTC1  | DMTC1 | CTC1  |  ---  |
 01 | *1    |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 10 | *2    | *3    |  ---  |  ---  | *4    | *5    |  ---  |  ---  |
 11 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
     *1 = BC instructions, see BC1 list
     *2 = S instr, see FPU list            *3 = D instr, see FPU list
     *4 = W instr, see FPU list            *5 = L instr, see FPU list
*/
RETURN_TYPE (*rc_COP1_Instruction[32])(OP_PARAMS)=
{
    rc_r4300i_COP1_mfc1,    rc_r4300i_COP1_dmfc1,   rc_r4300i_COP1_cfc1,    rc_UNUSED,  rc_r4300i_COP1_mtc1,    rc_r4300i_COP1_dmtc1,   rc_r4300i_COP1_ctc1,    rc_UNUSED,
    rc_COP1_BC_instr,       rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,
    rc_COP1_S_instr,        rc_COP1_D_instr,        rc_UNUSED,              rc_UNUSED,  rc_COP1_W_instr,        rc_COP1_L_instr,        rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED
};

//---------------------------------------------------------------------------------------

/*
    BC1: Instructions encoded by the nd and tf fields when opcode
         = COP1 and fmt = BC
    31--------26-25------21 ---17--16-------------------------------0
    |  = COP1   |  = BC   |    |nd|tf|                              |
    ------6----------5-----------1--1--------------------------------
    |---0---|---1---| tf
  0 | BC1F  | BC1T  |
  1 | BC1FL | BC1TL |
 nd |-------|-------|
*/
RETURN_TYPE (*rc_COP1_BC_Instruction[4])(OP_PARAMS)=
{
    rc_r4300i_COP1_bc1f,    rc_r4300i_COP1_bc1t,
    rc_r4300i_COP1_bc1fl,   rc_r4300i_COP1_bc1tl
};

//---------------------------------------------------------------------------------------

/*
    FPU: Instructions encoded by the function field when opcode = COP1
         and fmt = S
    31--------26-25------21 -------------------------------5--------0
    |  = COP1   |  = S    |                               | function|
    ------6----------5-----------------------------------------6-----
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 | ADD   | SUB   | MUL   | DIV   | SQRT  | ABS   | MOV   | NEG   |
001 |ROUND.L|TRUNC.L| CEIL.L|FLOOR.L|ROUND.W|TRUNC.W| CEIL.W|FLOOR.W|
010 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
011 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
100 |  ---  | CVT.D |  ---  |  ---  | CVT.W | CVT.L |  ---  |  ---  |
101 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
110 | C.F   | C.UN  | C.EQ  | C.UEQ | C.OLT | C.ULT | C.OLE | C.ULE |
111 | C.SF  | C.NGLE| C.SEQ | C.NGL | C.LT  | C.NGE | C.LE  | C.NGT |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
*/
RETURN_TYPE (*rc_COP1_S_Instruction[64])(OP_PARAMS)=
{
    rc_r4300i_COP1_add_s,   rc_r4300i_COP1_sub_s,   rc_r4300i_COP1_mul_s,   rc_r4300i_COP1_div_s,   rc_r4300i_COP1_sqrt_s,  rc_r4300i_COP1_abs_s,       rc_r4300i_COP1_mov_s,   rc_r4300i_COP1_neg_s,
    rc_r4300i_COP1_roundl,  rc_r4300i_COP1_truncl,  rc_r4300i_COP1_ceill,   rc_r4300i_COP1_floorl,  rc_r4300i_COP1_roundw,  rc_r4300i_COP1_truncw_s,    rc_r4300i_COP1_ceilw,   rc_r4300i_COP1_floorw,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,                  rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,                  rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,              rc_r4300i_COP1_cvtd_s,  rc_UNUSED,              rc_UNUSED,              rc_r4300i_COP1_cvtw_s,  rc_r4300i_COP1_cvtl_s,      rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,                  rc_UNUSED,              rc_UNUSED,
    rc_r4300i_C_F_S,        rc_r4300i_C_UN_S,       rc_r4300i_C_EQ_S,       rc_r4300i_C_UEQ_S,      rc_r4300i_C_OLT_S,      rc_r4300i_C_ULT_S,          rc_r4300i_C_OLE_S,      rc_r4300i_C_ULE_S,
    rc_r4300i_C_SF_S,       rc_r4300i_C_NGLE_S, rc_r4300i_C_SEQ_S,      rc_r4300i_C_NGL_S,      rc_r4300i_C_LT_S,       rc_r4300i_C_NGE_S,          rc_r4300i_C_LE_S,       rc_r4300i_C_NGT_S
};

//---------------------------------------------------------------------------------------

/*
    FPU: Instructions encoded by the function field when opcode = COP1
         and fmt = D
    31--------26-25------21 -------------------------------5--------0
    |  = COP1   |  = D    |                               | function|
    ------6----------5-----------------------------------------6-----
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 | ADD   | SUB   | MUL   | DIV   | SQRT  | ABS   | MOV   | NEG   |
001 |ROUND.L|TRUNC.L| CEIL.L|FLOOR.L|ROUND.W|TRUNC.W| CEIL.W|FLOOR.W|
010 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
011 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
100 | CVT.S |  ---  |  ---  |  ---  | CVT.W | CVT.L |  ---  |  ---  |
101 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
110 | C.F   | C.UN  | C.EQ  | C.UEQ | C.OLT | C.ULT | C.OLE | C.ULE |
111 | C.SF  | C.NGLE| C.SEQ | C.NGL | C.LT  | C.NGE | C.LE  | C.NGT |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
*/
RETURN_TYPE (*rc_COP1_D_Instruction[64])(OP_PARAMS)=
{
    rc_r4300i_COP1_add_d,   rc_r4300i_COP1_sub_d,   rc_r4300i_COP1_mul_d,   rc_r4300i_COP1_div_d,   rc_r4300i_COP1_sqrt_d,  rc_r4300i_COP1_abs_d,       rc_r4300i_COP1_mov_d,   rc_r4300i_COP1_neg_d,
    rc_r4300i_COP1_roundl,  rc_r4300i_COP1_truncl,  rc_r4300i_COP1_ceill,   rc_r4300i_COP1_floorl,  rc_r4300i_COP1_roundw,  rc_r4300i_COP1_truncw_d,    rc_r4300i_COP1_ceilw,   rc_r4300i_COP1_floorw,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,                  rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,                  rc_UNUSED,              rc_UNUSED,
    rc_r4300i_COP1_cvts_d,  rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_r4300i_COP1_cvtw_d,  rc_r4300i_COP1_cvtl_d,      rc_UNUSED,              rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,              rc_UNUSED,                  rc_UNUSED,              rc_UNUSED,
    rc_r4300i_C_F_D,        rc_r4300i_C_UN_D,       rc_r4300i_C_EQ_D,       rc_r4300i_C_UEQ_D,      rc_r4300i_C_OLT_D,      rc_r4300i_C_ULT_D,          rc_r4300i_C_OLE_D,      rc_r4300i_C_ULE_D,
    rc_r4300i_C_SF_D,       rc_r4300i_C_NGLE_D, rc_r4300i_C_SEQ_D,      rc_r4300i_C_NGL_D,      rc_r4300i_C_LT_D,       rc_r4300i_C_NGE_D,          rc_r4300i_C_LE_D,       rc_r4300i_C_NGT_D
};

//---------------------------------------------------------------------------------------

/*
    FPU: Instructions encoded by the function field when opcode = COP1
         and fmt = W
    31--------26-25------21 -------------------------------5--------0
    |  = COP1   |  = W    |                               | function|
    ------6----------5-----------------------------------------6-----
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
001 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
010 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
011 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
100 | CVT.S | CVT.D |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
101 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
110 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
111 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
*/
RETURN_TYPE (*rc_COP1_W_Instruction[64])(OP_PARAMS)=
{
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_r4300i_COP1_cvts_w,  rc_r4300i_COP1_cvtd_w,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED
};

//---------------------------------------------------------------------------------------

/*
    FPU: Instructions encoded by the function field when opcode = COP1
         and fmt = L
    31--------26-25------21 -------------------------------5--------0
    |  = COP1   |  = L    |                               | function|
    ------6----------5-----------------------------------------6-----
    |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
000 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
001 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
010 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
011 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
100 | CVT.S | CVT.D |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
101 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
110 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
111 |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |
 hi |-------|-------|-------|-------|-------|-------|-------|-------|
*/
RETURN_TYPE (*rc_COP1_L_Instruction[64])(OP_PARAMS)=
{
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_r4300i_COP1_cvts_l,  rc_r4300i_COP1_cvtd_l,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,
    rc_UNUSED,              rc_UNUSED,              rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED,  rc_UNUSED
};

//---------------------------------------------------------------------------------------

RETURN_TYPE rc_SPECIAL_instr(OP_PARAMS){    Dest = rc_SPECIAL_Instruction[_FUNCTION_](PASS_PARAMS);     return(Dest);}
RETURN_TYPE rc_REGIMM_instr(OP_PARAMS)  {   Dest = rc_REGIMM_Instruction[RT_FT](PASS_PARAMS);               return(Dest);}
RETURN_TYPE rc_COP0_instr(OP_PARAMS)    {   Dest = rc_COP0_Instruction[RS_BASE_FMT](PASS_PARAMS);           return(Dest);}
RETURN_TYPE rc_COP1_instr(OP_PARAMS)    {   Dest = rc_COP1_Instruction[RS_BASE_FMT](PASS_PARAMS);           return(Dest);}
RETURN_TYPE rc_TLB_instr(OP_PARAMS) {   Dest = rc_TLB_Instruction[_FUNCTION_](PASS_PARAMS);         return(Dest);}
RETURN_TYPE rc_COP1_BC_instr(OP_PARAMS){    Dest = rc_COP1_BC_Instruction[(_ND_ << 1) | _TF_](PASS_PARAMS);return(Dest);}
RETURN_TYPE rc_COP1_S_instr(OP_PARAMS)  {   Dest = rc_COP1_S_Instruction[_FUNCTION_](PASS_PARAMS);          return(Dest);}
RETURN_TYPE rc_COP1_D_instr(OP_PARAMS)  {   Dest = rc_COP1_D_Instruction[_FUNCTION_](PASS_PARAMS);          return(Dest);}
RETURN_TYPE rc_COP1_W_instr(OP_PARAMS)  {   Dest = rc_COP1_W_Instruction[_FUNCTION_](PASS_PARAMS);          return(Dest);}
RETURN_TYPE rc_COP1_L_instr(OP_PARAMS)  {   Dest = rc_COP1_L_Instruction[_FUNCTION_](PASS_PARAMS);          return(Dest);}